import logging
from typing import List, Tuple, Set, Type, Iterable, Any

from betty.ancestry import Person, Presence, Event, Subject, EventType, EVENT_TYPE_TYPES, DerivableEventType, \
    CreatableDerivableEventType, Ancestry
from betty.locale import DateRange, Date
from betty.parse import PostParser
from betty.plugin import Plugin, NO_CONFIGURATION
from betty.plugin.privatizer import Privatizer
from betty.site import Site


class DerivedEvent(Event):
    pass  # pragma: no cover


class DerivedDate(Date):
    @classmethod
    def derive(cls, date: Date) -> 'DerivedDate':
        return cls(date.year, date.month, date.day, fuzzy=date.fuzzy)


class Deriver(Plugin, PostParser):
    def __init__(self, ancestry: Ancestry):
        self._ancestry = ancestry

    @classmethod
    def for_site(cls, site: Site, configuration: Any = NO_CONFIGURATION):
        return cls(site.ancestry)

    async def post_parse(self,) -> None:
        await self.derive(self._ancestry)

    async def derive(self, ancestry: Ancestry) -> None:
        logger = logging.getLogger()
        for event_type_type in EVENT_TYPE_TYPES:
            event_type = event_type_type()
            if isinstance(event_type, DerivableEventType):
                created_derivations = 0
                updated_derivations = 0
                for person in ancestry.people.values():
                    created, updated = derive(person, event_type_type)
                    created_derivations += created
                    updated_derivations += updated
                logger.info('Updated %d %s events based on existing information.' % (updated_derivations, event_type.label))
                if isinstance(event_type, CreatableDerivableEventType):
                    logger.info('Created %d additional %s events based on existing information.' % (created_derivations, event_type.label))

    @classmethod
    def comes_before(cls) -> Set[Type[Plugin]]:
        return {Privatizer}


class _DateDeriver:
    @classmethod
    def derive(cls, person: Person, derivable_event: Event, reference_event_type_types: Set[Type[EventType]]) -> bool:
        if not reference_event_type_types:
            return False

        reference_events = _get_reference_events(person, reference_event_type_types)
        reference_events_dates = cls._get_events_dates(reference_events)
        reference_events_dates = filter(lambda x: x[1].comparable, reference_events_dates)
        if derivable_event.date is not None:
            reference_events_dates = filter(lambda x: cls._compare(derivable_event.date, x[1]), reference_events_dates)
        reference_events_dates = cls._sort(reference_events_dates)
        try:
            reference_event, reference_date = reference_events_dates[0]
        except IndexError:
            return False

        if derivable_event.date is None:
            derivable_event.date = DateRange()
        cls._set(derivable_event, DerivedDate.derive(reference_date))
        derivable_event.citations.append(*reference_event.citations)

        return True

    @classmethod
    def _get_events_dates(cls, events: Iterable[Event]) -> Iterable[Tuple[Event, Date]]:
        for event in events:
            if isinstance(event.date, Date):
                yield event, event.date
            if isinstance(event.date, DateRange):
                for date in cls._get_date_range_dates(event.date):
                    yield event, date

    @staticmethod
    def _get_date_range_dates(date: DateRange) -> Iterable[Date]:
        raise NotImplementedError

    @staticmethod
    def _compare(derivable_date: DateRange, reference_date: Date) -> bool:
        raise NotImplementedError

    @staticmethod
    def _sort(events_dates: Iterable[Tuple[Event, Date]]) -> List[Tuple[Event, Date]]:
        raise NotImplementedError

    @staticmethod
    def _set(derivable_event: Event, date: DerivedDate) -> None:
        raise NotImplementedError


class _ComesBeforeDateDeriver(_DateDeriver):
    @staticmethod
    def _get_date_range_dates(date: DateRange) -> Iterable[Date]:
        if date.start is not None and not date.start_is_boundary:
            yield date.start
        if date.end is not None:
            yield date.end

    @staticmethod
    def _compare(derivable_date: DateRange, reference_date: Date) -> bool:
        return derivable_date < reference_date

    @staticmethod
    def _sort(events_dates: Iterable[Tuple[Event, Date]]) -> List[Tuple[Event, Date]]:
        return sorted(events_dates, key=lambda x: x[1])

    @staticmethod
    def _set(derivable_event: Event, date: DerivedDate) -> None:
        derivable_event.date.end = date
        derivable_event.date.end_is_boundary = True


class _ComesAfterDateDeriver(_DateDeriver):
    @staticmethod
    def _get_date_range_dates(date: DateRange) -> Iterable[Date]:
        if date.start is not None:
            yield date.start
        if date.end is not None and not date.end_is_boundary:
            yield date.end

    @staticmethod
    def _compare(derivable_date: DateRange, reference_date: Date) -> bool:
        return derivable_date > reference_date

    @staticmethod
    def _sort(events_dates: Iterable[Tuple[Event, Date]]) -> List[Tuple[Event, Date]]:
        return sorted(events_dates, key=lambda x: x[1], reverse=True)

    @staticmethod
    def _set(derivable_event: Event, date: DerivedDate) -> None:
        derivable_event.date.start = date
        derivable_event.date.start_is_boundary = True


def derive(person: Person, event_type_type: Type[DerivableEventType]) -> Tuple[int, int]:
    # Gather any existing events that could be derived, or create a new derived event if needed.
    derivable_events = list(_get_derivable_events(person, event_type_type))
    if not derivable_events:
        if list(filter(lambda x: isinstance(x.event.type, event_type_type), person.presences)):
            return 0, 0
        if issubclass(event_type_type, CreatableDerivableEventType):
            derivable_events = [DerivedEvent(event_type_type())]
        else:
            return 0, 0

    # Aggregate event type order from references and backreferences.
    comes_before_event_type_types = event_type_type.comes_before()
    comes_after_event_type_types = event_type_type.comes_after()
    for other_event_type_type in EVENT_TYPE_TYPES:
        if event_type_type in other_event_type_type.comes_before():
            comes_after_event_type_types.add(other_event_type_type)
        if event_type_type in other_event_type_type.comes_after():
            comes_before_event_type_types.add(other_event_type_type)

    created_derivations = 0
    updated_derivations = 0

    for derivable_event in derivable_events:
        dates_derived = False

        if derivable_event.date is None or derivable_event.date.end is None:
            dates_derived = dates_derived or _ComesBeforeDateDeriver.derive(person, derivable_event, comes_before_event_type_types)

        if derivable_event.date is None or derivable_event.date.start is None:
            dates_derived = dates_derived or _ComesAfterDateDeriver.derive(person, derivable_event, comes_after_event_type_types)

        if dates_derived:
            if isinstance(derivable_event, DerivedEvent):
                created_derivations += 1
                Presence(person, Subject(), derivable_event)
            else:
                updated_derivations += 1

    return created_derivations, updated_derivations


def _get_derivable_events(person: Person, derivable_event_type_type: Type[EventType]) -> Iterable[Event]:
    for presence in person.presences:
        event = presence.event

        # Ignore events that have been derived already.
        if isinstance(event, DerivedEvent):
            continue

        # Ignore events of the wrong type.
        if not isinstance(event.type, derivable_event_type_type):
            continue

        # Ignore events with enough date information that nothing more can be derived.
        if isinstance(event.date, Date):
            continue
        if isinstance(event.date, DateRange) and (not event.type.comes_after() or event.date.start is not None) and (not event.type.comes_before() or event.date.end is not None):
            continue

        yield presence.event


def _get_reference_events(person: Person, reference_event_type_types: Set[Type[EventType]]) -> Iterable[Event]:
    for reference_event in [presence.event for presence in person.presences]:
        # We cannot reliably determine dates based on reference events with calculated date ranges, as those events
        # would start or end *sometime* during the date range, but to derive dates we need reference events' exact
        # start and end dates.
        if isinstance(reference_event, DerivedEvent):
            continue

        # Ignore reference events of the wrong type.
        if not isinstance(reference_event.type, tuple(reference_event_type_types)):
            continue

        yield reference_event



from tempfile import TemporaryDirectory

from betty.ancestry import Ancestry, Person, Place, Presence, PlaceName, IdentifiableEvent, File, PersonName, \
    IdentifiableSource, IdentifiableCitation, Subject, Birth, Enclosure
from betty.config import Configuration
from betty.asyncio import sync
from betty.parse import parse
from betty.plugin.cleaner import Cleaner, clean
from betty.site import Site
from betty.tests import TestCase


class CleanerTest(TestCase):
    @sync
    async def test_post_parse(self) -> None:
        event = IdentifiableEvent('E0', Birth())
        with TemporaryDirectory() as output_directory_path:
            configuration = Configuration(
                output_directory_path, 'https://example.com')
            configuration.plugins[Cleaner] = None
            async with Site(configuration) as site:
                site.ancestry.events[event.id] = event
                await parse(site)
                self.assertEquals({}, site.ancestry.events)


class CleanTest(TestCase):
    def test_clean(self) -> None:
        ancestry = Ancestry()

        onymous_event = IdentifiableEvent('E0', Birth())
        Presence(Person('P0'), Subject(), onymous_event)
        ancestry.events[onymous_event.id] = onymous_event

        anonymous_event = IdentifiableEvent('E1', Birth())
        ancestry.events[anonymous_event.id] = anonymous_event

        onymous_place = Place('P0', [PlaceName('Amsterdam')])
        onymous_place.events.append(onymous_event)
        ancestry.places[onymous_place.id] = onymous_place

        anonymous_place = Place('P1', [PlaceName('Almelo')])
        ancestry.places[anonymous_place.id] = anonymous_place

        onmyous_place_because_encloses_onmyous_places = Place(
            'P3', [PlaceName('Netherlands')])
        Enclosure(onymous_place, onmyous_place_because_encloses_onmyous_places)
        Enclosure(anonymous_place, onmyous_place_because_encloses_onmyous_places)
        ancestry.places[
            onmyous_place_because_encloses_onmyous_places.id] = onmyous_place_because_encloses_onmyous_places

        clean(ancestry)

        self.assertDictEqual({
            onymous_event.id: onymous_event,
        }, ancestry.events)
        self.assertDictEqual({
            onymous_place.id: onymous_place,
            onmyous_place_because_encloses_onmyous_places.id: onmyous_place_because_encloses_onmyous_places,
        }, ancestry.places)

        self.assertNotIn(
            anonymous_place, onmyous_place_because_encloses_onmyous_places.encloses)

    def test_clean_should_not_clean_person_if_public(self):
        ancestry = Ancestry()

        person = Person('P0')
        person.private = False
        ancestry.people[person.id] = person

        clean(ancestry)

        self.assertEqual(person, ancestry.people[person.id])

    def test_clean_should_clean_person_with_private_children(self) -> None:
        ancestry = Ancestry()

        person = Person('P0')
        person.private = True
        ancestry.people[person.id] = person
        child = Person('P1')
        child.private = True
        ancestry.people[child.id] = child
        grandchild = Person('P2')
        grandchild.private = True
        ancestry.people[grandchild.id] = grandchild
        great_grandchild = Person('P3')
        great_grandchild.private = True
        ancestry.people[great_grandchild.id] = great_grandchild

        clean(ancestry)

        self.assertNotIn(person.id, ancestry.people)

    def test_clean_should_not_clean_person_with_public_children(self):
        ancestry = Ancestry()

        person = Person('P0')
        person.private = False
        ancestry.people[person.id] = person
        child = Person('P1')
        child.private = True
        ancestry.people[child.id] = child
        grandchild = Person('P2')
        grandchild.private = True
        ancestry.people[grandchild.id] = grandchild
        great_grandchild = Person('P3')
        great_grandchild.private = False
        ancestry.people[great_grandchild.id] = great_grandchild

        clean(ancestry)

        self.assertEqual(person, ancestry.people[person.id])

    def test_clean_should_clean_event(self) -> None:
        ancestry = Ancestry()

        source = IdentifiableSource('S1', 'The Source')
        ancestry.sources[source.id] = source

        citation = IdentifiableCitation('C1', source)
        ancestry.citations[citation.id] = citation

        file = File('F1', __file__)
        ancestry.files[file.id] = file

        place = Place('P0', [PlaceName('The Place')])
        ancestry.places[place.id] = place

        event = IdentifiableEvent('E0', Birth())
        event.citations.append(citation)
        event.files.append(file)
        event.place = place
        ancestry.events[event.id] = event

        clean(ancestry)

        self.assertNotIn(event.id, ancestry.events)
        self.assertIsNone(event.place)
        self.assertNotIn(event, place.events)
        self.assertNotIn(place.id, ancestry.places)
        self.assertNotIn(event, citation.facts)
        self.assertNotIn(citation.id, ancestry.citations)
        self.assertNotIn(event, file.resources)
        self.assertNotIn(file.id, ancestry.files)

    def test_clean_should_not_clean_event_with_presences_with_people(self) -> None:
        ancestry = Ancestry()

        source = IdentifiableSource('S1', 'The Source')
        ancestry.sources[source.id] = source

        citation = IdentifiableCitation('C1', source)
        ancestry.citations[citation.id] = citation

        file = File('F1', __file__)
        ancestry.files[file.id] = file

        place = Place('P0', [PlaceName('The Place')])
        ancestry.places[place.id] = place

        person = Person('P0')

        event = IdentifiableEvent('E0', Birth())
        event.citations.append(citation)
        event.files.append(file)
        event.place = place
        ancestry.events[event.id] = event

        Presence(person, Subject(), event)

        clean(ancestry)

        self.assertEqual(event, ancestry.events[event.id])
        self.assertIn(event, place.events)
        self.assertEqual(place, ancestry.places[place.id])
        self.assertIn(event, citation.facts)
        self.assertEqual(citation, ancestry.citations[citation.id])
        self.assertIn(event, file.resources)
        self.assertEqual(file, ancestry.files[file.id])

    def test_clean_should_clean_file(self) -> None:
        ancestry = Ancestry()

        file = File('F0', __file__)
        ancestry.files[file.id] = file

        clean(ancestry)

        self.assertNotIn(file.id, ancestry.files)

    def test_clean_should_not_clean_file_with_resources(self) -> None:
        ancestry = Ancestry()

        person = Person('P0')
        ancestry.people[person.id] = person

        file = File('F0', __file__)
        file.resources.append(person)
        ancestry.files[file.id] = file

        clean(ancestry)

        self.assertEqual(file, ancestry.files[file.id])
        self.assertIn(person, file.resources)
        self.assertEqual(person, ancestry.people[person.id])

    def test_clean_should_clean_source(self) -> None:
        ancestry = Ancestry()

        source = IdentifiableSource('S0', 'The source')
        ancestry.sources[source.id] = source

        clean(ancestry)

        self.assertNotIn(source.id, ancestry.sources)

    def test_clean_should_not_clean_source_with_citations(self) -> None:
        ancestry = Ancestry()

        source = IdentifiableSource('S0', 'The Source')
        ancestry.sources[source.id] = source

        citation = IdentifiableCitation('C0', source)
        citation.facts.append(PersonName('Jane'))
        ancestry.citations[citation.id] = citation

        clean(ancestry)

        self.assertEqual(source, ancestry.sources[source.id])
        self.assertEqual(source, citation.source)
        self.assertEqual(citation, ancestry.citations[citation.id])

    def test_clean_should_not_clean_source_with_contained_by(self) -> None:
        ancestry = Ancestry()

        source = IdentifiableSource('S0', 'The Source')
        ancestry.sources[source.id] = source

        contained_by = IdentifiableSource('S1', 'The Source')
        contained_by.contains.append(source)
        ancestry.sources[contained_by.id] = contained_by

        clean(ancestry)

        self.assertEqual(source, ancestry.sources[source.id])
        self.assertIn(source, contained_by.contains)
        self.assertEqual(contained_by, ancestry.sources[contained_by.id])

    def test_clean_should_not_clean_source_with_contains(self) -> None:
        ancestry = Ancestry()

        source = IdentifiableSource('S0', 'The Source')
        ancestry.sources[source.id] = source

        contains = IdentifiableSource('S1', 'The Source')
        contains.contained_by = source
        ancestry.sources[contains.id] = contains

        clean(ancestry)

        self.assertEqual(source, ancestry.sources[source.id])
        self.assertEqual(source, contains.contained_by)
        self.assertEqual(contains, ancestry.sources[contains.id])

    def test_clean_should_not_clean_source_with_files(self) -> None:
        ancestry = Ancestry()

        source = IdentifiableSource('S0', 'The Source')
        ancestry.sources[source.id] = source

        file = File('F0', __file__)
        file.resources.append(source)
        ancestry.files[file.id] = file

        clean(ancestry)

        self.assertEqual(source, ancestry.sources[source.id])
        self.assertIn(source, file.sources)
        self.assertEqual(file, ancestry.files[file.id])

    def test_clean_should_clean_citation(self) -> None:
        ancestry = Ancestry()

        source = IdentifiableSource('S0', 'The source')
        ancestry.sources[source.id] = source

        citation = IdentifiableCitation('C0', source)
        ancestry.citations[citation.id] = citation

        clean(ancestry)

        self.assertNotIn(citation.id, ancestry.citations)
        self.assertNotIn(citation, source.citations)

    def test_clean_should_not_clean_citation_with_facts(self) -> None:
        ancestry = Ancestry()

        source = IdentifiableSource('S0', 'The Source')
        ancestry.sources[source.id] = source

        citation = IdentifiableCitation('C0', source)
        citation.facts.append(PersonName('Jane'))
        ancestry.citations[citation.id] = citation

        fact = Person('P0')
        fact.citations.append(citation)
        ancestry.people[fact.id] = fact

        clean(ancestry)

        self.assertEqual(citation, ancestry.citations[citation.id])
        self.assertIn(citation, fact.citations)
        self.assertEqual(fact, ancestry.people[fact.id])

    def test_clean_should_not_clean_citation_with_files(self) -> None:
        ancestry = Ancestry()

        source = IdentifiableSource('S0', 'The Source')
        ancestry.sources[source.id] = source

        citation = IdentifiableCitation('C0', source)
        ancestry.citations[citation.id] = citation

        file = File('F0', __file__)
        file.resources.append(citation)
        ancestry.files[file.id] = file

        clean(ancestry)

        self.assertEqual(citation, ancestry.citations[citation.id])
        self.assertIn(citation, file.citations)
        self.assertEqual(file, ancestry.files[file.id])
        self.assertIn(citation, source.citations)
        self.assertEqual(source, ancestry.sources[source.id])



import os
from glob import glob
from typing import Any

import sass

from betty.render import Renderer


class SassRenderer(Renderer):
    _EXTENSIONS = ('sass', 'scss')

    async def render_file(self, file_path: str) -> None:
        for extension in self._EXTENSIONS:
            if file_path.endswith(self._EXTENSIONS):
                css_file_path = '%scss' % file_path[:-len(extension)]
                break
        else:
            return
        with open(css_file_path, 'w') as f:
            f.write(self._compile(filename=file_path))
        os.remove(file_path)

    async def render_tree(self, tree_path: str) -> None:
        self._compile(dirname=(tree_path, tree_path))
        for extension in self._EXTENSIONS:
            patterns = [
                # Files in the path.
                os.path.join(tree_path, '*.' + extension),
                # Files in the path's subdirectories.
                os.path.join(tree_path, '**', '*.' + extension),
            ]
            for pattern in patterns:
                for file_path in glob(pattern):
                    os.remove(file_path)

    def _compile(self, **kwargs) -> Any:
        return sass.compile(output_style='compressed', **kwargs)



import json as stdjson
from os import makedirs, path
from os.path import join, exists
from tempfile import TemporaryDirectory, NamedTemporaryFile

import html5lib
from lxml import etree

from betty import json
from betty.ancestry import Person, Place, Source, PlaceName, File, IdentifiableEvent, IdentifiableCitation, \
    IdentifiableSource, Birth
from betty.config import Configuration, LocaleConfiguration
from betty.asyncio import sync
from betty.generate import generate
from betty.site import Site
from betty.tests import TestCase


class GenerateTestCase(TestCase):
    def setUp(self):
        self._outputDirectory = TemporaryDirectory()
        self.site = None

    def tearDown(self):
        self._outputDirectory.cleanup()

    def assert_betty_html(self, path: str) -> str:
        file_path = join(
            self.site.configuration.www_directory_path, path.lstrip('/'))
        self.assertTrue(exists(file_path), '%s does not exist' % file_path)
        with open(file_path) as f:
            parser = html5lib.HTMLParser(strict=True)
            parser.parse(f)
        return file_path

    def assert_betty_json(self, path: str, schema_definition: str) -> str:
        file_path = join(
            self.site.configuration.www_directory_path, path.lstrip('/'))
        self.assertTrue(exists(file_path), '%s does not exist' % file_path)
        with open(file_path) as f:
            json.validate(stdjson.load(f), schema_definition, self.site)
        return file_path


class RenderTest(GenerateTestCase):
    def setUp(self):
        GenerateTestCase.setUp(self)
        configuration = Configuration(
            self._outputDirectory.name, 'https://ancestry.example.com')
        self.site = Site(configuration)

    @sync
    async def test_front_page(self):
        await generate(self.site)
        self.assert_betty_html('/index.html')

    @sync
    async def test_files(self):
        await generate(self.site)
        self.assert_betty_html('/file/index.html')
        self.assert_betty_json('/file/index.json', 'fileCollection')

    @sync
    async def test_file(self):
        with NamedTemporaryFile() as f:
            file = File('PLACE1', f.name)
            self.site.ancestry.files[file.id] = file
            await generate(self.site)
            self.assert_betty_html('/file/%s/index.html' % file.id)
            self.assert_betty_json('/file/%s/index.json' % file.id, 'file')

    @sync
    async def test_places(self):
        await generate(self.site)
        self.assert_betty_html('/place/index.html')
        self.assert_betty_json('/place/index.json', 'placeCollection')

    @sync
    async def test_place(self):
        place = Place('PLACE1', [PlaceName('one')])
        self.site.ancestry.places[place.id] = place
        await generate(self.site)
        self.assert_betty_html('/place/%s/index.html' % place.id)
        self.assert_betty_json('/place/%s/index.json' % place.id, 'place')

    @sync
    async def test_people(self):
        await generate(self.site)
        self.assert_betty_html('/person/index.html')
        self.assert_betty_json('/person/index.json', 'personCollection')

    @sync
    async def test_person(self):
        person = Person('PERSON1')
        self.site.ancestry.people[person.id] = person
        await generate(self.site)
        self.assert_betty_html('/person/%s/index.html' % person.id)
        self.assert_betty_json('/person/%s/index.json' % person.id, 'person')

    @sync
    async def test_events(self):
        await generate(self.site)
        self.assert_betty_html('/event/index.html')
        self.assert_betty_json('/event/index.json', 'eventCollection')

    @sync
    async def test_event(self):
        event = IdentifiableEvent('EVENT1', Birth())
        self.site.ancestry.events[event.id] = event
        await generate(self.site)
        self.assert_betty_html('/event/%s/index.html' % event.id)
        self.assert_betty_json('/event/%s/index.json' % event.id, 'event')

    @sync
    async def test_citation(self):
        citation = IdentifiableCitation('CITATION1', Source('A Little Birdie'))
        self.site.ancestry.citations[citation.id] = citation
        await generate(self.site)
        self.assert_betty_html('/citation/%s/index.html' % citation.id)
        self.assert_betty_json('/citation/%s/index.json' %
                               citation.id, 'citation')

    @sync
    async def test_sources(self):
        await generate(self.site)
        self.assert_betty_html('/source/index.html')
        self.assert_betty_json('/source/index.json', 'sourceCollection')

    @sync
    async def test_source(self):
        source = IdentifiableSource('SOURCE1', 'A Little Birdie')
        self.site.ancestry.sources[source.id] = source
        await generate(self.site)
        self.assert_betty_html('/source/%s/index.html' % source.id)
        self.assert_betty_json('/source/%s/index.json' % source.id, 'source')


class MultilingualTest(GenerateTestCase):
    def setUp(self):
        GenerateTestCase.setUp(self)
        configuration = Configuration(
            self._outputDirectory.name, 'https://ancestry.example.com')
        configuration.locales.clear()
        configuration.locales['nl'] = LocaleConfiguration('nl')
        configuration.locales['en'] = LocaleConfiguration('en')
        self.site = Site(configuration)

    @sync
    async def test_root_redirect(self):
        await generate(self.site)
        with open(self.assert_betty_html('/index.html')) as f:
            meta_redirect = '<meta http-equiv="refresh" content="0; url=/nl/index.html">'
            self.assertIn(meta_redirect, f.read())

    @sync
    async def test_public_localized_resource(self):
        await generate(self.site)
        with open(self.assert_betty_html('/nl/index.html')) as f:
            translation_link = '<a href="/en/index.html" hreflang="en" lang="en" rel="alternate">English</a>'
            self.assertIn(translation_link, f.read())
        with open(self.assert_betty_html('/en/index.html')) as f:
            translation_link = '<a href="/nl/index.html" hreflang="nl" lang="nl" rel="alternate">Nederlands</a>'
            self.assertIn(translation_link, f.read())

    @sync
    async def test_entity(self):
        person = Person('PERSON1')
        self.site.ancestry.people[person.id] = person
        await generate(self.site)
        with open(self.assert_betty_html('/nl/person/%s/index.html' % person.id)) as f:
            translation_link = '<a href="/en/person/%s/index.html" hreflang="en" lang="en" rel="alternate">English</a>' % person.id
            self.assertIn(translation_link, f.read())
        with open(self.assert_betty_html('/en/person/%s/index.html' % person.id)) as f:
            translation_link = '<a href="/nl/person/%s/index.html" hreflang="nl" lang="nl" rel="alternate">Nederlands</a>' % person.id
            self.assertIn(translation_link, f.read())


class ResourceOverrideTest(GenerateTestCase):
    @sync
    async def test(self):
        with TemporaryDirectory() as output_directory_path:
            with TemporaryDirectory() as assets_directory_path:
                makedirs(join(assets_directory_path, 'public', 'localized'))
                with open(join(assets_directory_path, 'public', 'localized', 'index.html.j2'), 'w') as f:
                    f.write('{% block page_content %}Betty was here{% endblock %}')
                configuration = Configuration(
                    output_directory_path, 'https://ancestry.example.com')
                configuration.assets_directory_path = assets_directory_path
                site = Site(configuration)
                await generate(site)
                with open(join(configuration.www_directory_path, 'index.html')) as f:
                    self.assertIn('Betty was here', f.read())


class SitemapRenderTest(GenerateTestCase):
    def setUp(self):
        GenerateTestCase.setUp(self)
        configuration = Configuration(
            self._outputDirectory.name, 'https://ancestry.example.com')
        self.site = Site(configuration)

    @sync
    async def test_validate(self):
        await generate(self.site)
        with open(path.join(path.dirname(__file__), 'test_generate_assets', 'sitemap.xsd')) as f:
            schema_doc = etree.parse(f)
        schema = etree.XMLSchema(schema_doc)
        with open(path.join(self.site.configuration.www_directory_path, 'sitemap.xml')) as f:
            sitemap_doc = etree.parse(f)
        schema.validate(sitemap_doc)



from os import path
from shutil import copyfile

from jinja2 import FileSystemLoader, Environment

from betty.fs import makedirs


async def generate_configuration_file(destination_file_path: str, jinja2_environment: Environment, **kwargs) -> None:
    configuration_file_template_path = path.join(path.dirname(__file__), 'assets', 'nginx.conf.j2')
    template = FileSystemLoader('/').load(jinja2_environment, configuration_file_template_path, jinja2_environment.globals)
    makedirs(path.dirname(destination_file_path))
    with open(destination_file_path, 'w') as f:
        f.write(await template.render_async(kwargs))


async def generate_dockerfile_file(destination_file_path: str) -> None:
    makedirs(path.dirname(destination_file_path))
    copyfile(path.join(path.dirname(__file__), 'assets', 'docker', 'Dockerfile'), destination_file_path)



from typing import List, Type, Set, Any

from voluptuous import Schema, Required

from betty.ancestry import Ancestry
from betty.config import Configuration
from betty.asyncio import sync
from betty.graph import CyclicGraphError
from betty.plugin import Plugin, NO_CONFIGURATION
from betty.site import Site
from betty.tests import TestCase


class Tracker:
    async def track(self, carrier: List):
        raise NotImplementedError


class TrackablePlugin(Plugin, Tracker):
    async def track(self, carrier: List):
        carrier.append(self)


class NonConfigurablePlugin(TrackablePlugin):
    pass  # pragma: no cover


class ConfigurablePlugin(Plugin):
    configuration_schema: Schema = Schema({
        Required('check'): lambda x: x
    })

    def __init__(self, check):
        self.check = check

    @classmethod
    def for_site(cls, site: Site, configuration: Any = NO_CONFIGURATION):
        return cls(configuration['check'])


class CyclicDependencyOnePlugin(Plugin):
    @classmethod
    def depends_on(cls) -> Set[Type[Plugin]]:
        return {CyclicDependencyTwoPlugin}


class CyclicDependencyTwoPlugin(Plugin):
    @classmethod
    def depends_on(cls) -> Set[Type[Plugin]]:
        return {CyclicDependencyOnePlugin}


class DependsOnNonConfigurablePluginPlugin(TrackablePlugin):
    @classmethod
    def depends_on(cls) -> Set[Type[Plugin]]:
        return {NonConfigurablePlugin}


class AlsoDependsOnNonConfigurablePluginPlugin(TrackablePlugin):
    @classmethod
    def depends_on(cls) -> Set[Type[Plugin]]:
        return {NonConfigurablePlugin}


class DependsOnNonConfigurablePluginPluginPlugin(TrackablePlugin):
    @classmethod
    def depends_on(cls) -> Set[Type[Plugin]]:
        return {DependsOnNonConfigurablePluginPlugin}


class ComesBeforeNonConfigurablePluginPlugin(TrackablePlugin):
    @classmethod
    def comes_before(cls) -> Set[Type[Plugin]]:
        return {NonConfigurablePlugin}


class ComesAfterNonConfigurablePluginPlugin(TrackablePlugin):
    @classmethod
    def comes_after(cls) -> Set[Type[Plugin]]:
        return {NonConfigurablePlugin}


class SiteTest(TestCase):
    _MINIMAL_CONFIGURATION_ARGS = {
        'output_directory_path': '/tmp/path/to/site',
        'base_url': 'https://example.com',
    }

    @sync
    async def test_ancestry_should_return(self):
        configuration = Configuration(**self._MINIMAL_CONFIGURATION_ARGS)
        async with Site(configuration) as sut:
            self.assertIsInstance(sut.ancestry, Ancestry)

    @sync
    async def test_configuration_should_return(self):
        configuration = Configuration(**self._MINIMAL_CONFIGURATION_ARGS)
        async with Site(configuration) as sut:
            self.assertEquals(configuration, sut.configuration)

    @sync
    async def test_with_one_plugin(self):
        configuration = Configuration(**self._MINIMAL_CONFIGURATION_ARGS)
        configuration.plugins[NonConfigurablePlugin] = None
        async with Site(configuration) as sut:
            self.assertEquals(1, len(sut.plugins))
            self.assertIsInstance(
                sut.plugins[NonConfigurablePlugin], NonConfigurablePlugin)

    @sync
    async def test_with_one_configurable_plugin(self):
        configuration = Configuration(**self._MINIMAL_CONFIGURATION_ARGS)
        check = 1337
        configuration.plugins[ConfigurablePlugin] = {
            'check': check,
        }
        async with Site(configuration) as sut:
            self.assertEquals(1, len(sut.plugins))
            self.assertIsInstance(
                sut.plugins[ConfigurablePlugin], ConfigurablePlugin)
            self.assertEquals(check, sut.plugins[ConfigurablePlugin].check)

    @sync
    async def test_with_one_plugin_with_single_chained_dependency(self):
        configuration = Configuration(**self._MINIMAL_CONFIGURATION_ARGS)
        configuration.plugins[DependsOnNonConfigurablePluginPluginPlugin] = None
        async with Site(configuration) as sut:
            carrier = []
            await sut.dispatcher.dispatch(Tracker, 'track')(carrier)
            self.assertEquals(3, len(carrier))
            self.assertEquals(NonConfigurablePlugin, type(carrier[0]))
            self.assertEquals(DependsOnNonConfigurablePluginPlugin,
                              type(carrier[1]))
            self.assertEquals(
                DependsOnNonConfigurablePluginPluginPlugin, type(carrier[2]))

    @sync
    async def test_with_multiple_plugins_with_duplicate_dependencies(self):
        configuration = Configuration(**self._MINIMAL_CONFIGURATION_ARGS)
        configuration.plugins[DependsOnNonConfigurablePluginPlugin] = None
        configuration.plugins[AlsoDependsOnNonConfigurablePluginPlugin] = None
        async with Site(configuration) as sut:
            carrier = []
            await sut.dispatcher.dispatch(Tracker, 'track')(carrier)
            self.assertEquals(3, len(carrier))
            self.assertEquals(NonConfigurablePlugin, type(carrier[0]))
            self.assertIn(DependsOnNonConfigurablePluginPlugin, [
                type(plugin) for plugin in carrier])
            self.assertIn(AlsoDependsOnNonConfigurablePluginPlugin, [
                type(plugin) for plugin in carrier])

    @sync
    async def test_with_multiple_plugins_with_cyclic_dependencies(self):
        configuration = Configuration(**self._MINIMAL_CONFIGURATION_ARGS)
        configuration.plugins[CyclicDependencyOnePlugin] = None
        with self.assertRaises(CyclicGraphError):
            async with Site(configuration):
                pass

    @sync
    async def test_with_comes_before_with_other_plugin(self):
        configuration = Configuration(**self._MINIMAL_CONFIGURATION_ARGS)
        configuration.plugins[NonConfigurablePlugin] = None
        configuration.plugins[ComesBeforeNonConfigurablePluginPlugin] = None
        async with Site(configuration) as sut:
            carrier = []
            await sut.dispatcher.dispatch(Tracker, 'track')(carrier)
            self.assertEquals(2, len(carrier))
            self.assertEquals(
                ComesBeforeNonConfigurablePluginPlugin, type(carrier[0]))
            self.assertEquals(NonConfigurablePlugin, type(carrier[1]))

    @sync
    async def test_with_comes_before_without_other_plugin(self):
        configuration = Configuration(**self._MINIMAL_CONFIGURATION_ARGS)
        configuration.plugins[ComesBeforeNonConfigurablePluginPlugin] = None
        async with Site(configuration) as sut:
            carrier = []
            await sut.dispatcher.dispatch(Tracker, 'track')(carrier)
            self.assertEquals(1, len(carrier))
            self.assertEquals(
                ComesBeforeNonConfigurablePluginPlugin, type(carrier[0]))

    @sync
    async def test_with_comes_after_with_other_plugin(self):
        configuration = Configuration(**self._MINIMAL_CONFIGURATION_ARGS)
        configuration.plugins[ComesAfterNonConfigurablePluginPlugin] = None
        configuration.plugins[NonConfigurablePlugin] = None
        async with Site(configuration) as sut:
            carrier = []
            await sut.dispatcher.dispatch(Tracker, 'track')(carrier)
            self.assertEquals(2, len(carrier))
            self.assertEquals(NonConfigurablePlugin, type(carrier[0]))
            self.assertEquals(ComesAfterNonConfigurablePluginPlugin,
                              type(carrier[1]))

    @sync
    async def test_with_comes_after_without_other_plugin(self):
        configuration = Configuration(**self._MINIMAL_CONFIGURATION_ARGS)
        configuration.plugins[ComesAfterNonConfigurablePluginPlugin] = None
        async with Site(configuration) as sut:
            carrier = []
            await sut.dispatcher.dispatch(Tracker, 'track')(carrier)
            self.assertEquals(1, len(carrier))
            self.assertEquals(ComesAfterNonConfigurablePluginPlugin,
                              type(carrier[0]))

    @sync
    async def test_resources_without_assets_directory_path(self):
        configuration = Configuration(**self._MINIMAL_CONFIGURATION_ARGS)
        async with Site(configuration) as sut:
            self.assertEquals(1, len(sut.assets.paths))

    @sync
    async def test_resources_with_assets_directory_path(self):
        assets_directory_path = '/tmp/betty'
        configuration = Configuration(**self._MINIMAL_CONFIGURATION_ARGS)
        configuration.assets_directory_path = assets_directory_path
        async with Site(configuration) as sut:
            self.assertEquals(2, len(sut.assets.paths))
            self.assertEquals(assets_directory_path, sut.assets.paths[0])



import logging
from contextlib import suppress
try:
    from contextlib import asynccontextmanager
except ImportError:
    from async_generator import asynccontextmanager
from tempfile import TemporaryDirectory
from typing import Optional, Dict, Callable, Tuple
import unittest

from jinja2 import Environment, Template

import betty
from betty.config import Configuration
from betty.site import Site


def patch_cache(f):
    def _patch_cache(*args, **kwargs):
        original_cache_directory_path = betty._CACHE_DIRECTORY_PATH
        cache_directory = TemporaryDirectory()
        betty._CACHE_DIRECTORY_PATH = cache_directory.name
        try:
            f(*args, **kwargs)
        finally:
            betty._CACHE_DIRECTORY_PATH = original_cache_directory_path
            # Pythons 3.6 and 3.7 do not allow the temporary directory to have been removed already.
            with suppress(FileNotFoundError):
                cache_directory.cleanup()

    return _patch_cache


class TestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        logging.disable(logging.CRITICAL)

    @classmethod
    def tearDownClass(cls) -> None:
        logging.disable(logging.NOTSET)


class TemplateTestCase(TestCase):
    template_string = None
    template_file = None

    def __init__(self, *args, **kwargs):
        TestCase.__init__(self, *args, **kwargs)
        if self.template_string is not None and self.template_file is not None:
            class_name = self.__class__.__name__
            raise RuntimeError(f'{class_name} must define either `{class_name}.template_string` or `{class_name}.template_file`, but not both.')

    @asynccontextmanager
    async def _render(self, data: Optional[Dict] = None, template_file: Optional[Template] = None, template_string: Optional[str] = None, update_configuration: Optional[Callable[[Configuration], None]] = None) -> Tuple[str, Site]:
        if template_string is not None and template_file is not None:
            raise RuntimeError('You must define either `template_string` or `template_file`, but not both.')
        if template_string is not None:
            template = template_string
            template_factory = Environment.from_string
        elif template_file is not None:
            template = template_file
            template_factory = Environment.get_template
        elif self.template_string is not None:
            template = self.template_string
            template_factory = Environment.from_string
        elif self.template_file is not None:
            template = self.template_file
            template_factory = Environment.get_template
        else:
            class_name = self.__class__.__name__
            raise RuntimeError(f'You must define one of `template_string`, `template_file`, `{class_name}.template_string`, or `{class_name}.template_file`.')
        if data is None:
            data = {}
        with TemporaryDirectory() as output_directory_path:
            configuration = Configuration(output_directory_path, 'https://example.com')
            configuration.mode = 'development'
            if update_configuration is not None:
                update_configuration(configuration)
            async with Site(configuration) as site:
                rendered = await template_factory(site.jinja2_environment, template).render_async(**data)
                # We want to keep the site around, but we must make sure all dispatched tasks are done, so we shut down
                # the executor. Crude, but effective.
                site.executor.shutdown()
                yield rendered, site



from parameterized import parameterized

from betty.ancestry import Place, PlaceName
from betty.asyncio import sync
from betty.locale import DateRange, Date
from betty.tests import TemplateTestCase


class Test(TemplateTestCase):
    template_file = 'label/place.html.j2'

    @parameterized.expand([
        ('<address><a href="/place/P0/index.html"><span>The Place</span></a></address>',
         {
             'place': Place('P0', [PlaceName('The Place')]),
         }),
        ('<address><a href="/place/P0/index.html"><span lang="en">The Place</span></a></address>',
         {
             'place': Place('P0', [PlaceName('The Place', 'en')]),
         }),
        ('<address><a href="/place/P0/index.html"><span lang="nl">De Plaats</span></a></address>',
         {
             'place': Place('P0', [PlaceName('The Place', 'en'), PlaceName('De Plaats', 'nl')]),
             'locale': 'nl',
         }),
        ('<address><span>The Place</span></address>',
         {
             'place': Place('P0', [PlaceName('The Place')]),
             'embedded': True,
         }),
        ('<address><a href="/place/P0/index.html"><span lang="nl">De Nieuwe Plaats</span></a></address>',
         {
             'place': Place('P0', [PlaceName('The Old Place', 'en', date=DateRange(None, Date(1969, 12, 31))),
                                   PlaceName('De Nieuwe Plaats', 'nl', date=DateRange(Date(1970, 1, 1)))]),
             'locale': 'nl',
             'date_context': Date(1970, 1, 1),
         })
    ])
    @sync
    async def test(self, expected, data):
        async with self._render(data=data) as (actual, _):
            self.assertEqual(expected, actual)



import logging
import sys
from logging import CRITICAL, ERROR, WARNING, INFO, DEBUG, NOTSET, StreamHandler


class CliHandler(StreamHandler):
    COLOR_LEVELS = [
        (CRITICAL, 91),
        (ERROR, 91),
        (WARNING, 93),
        (INFO, 92),
        (DEBUG, 97),
        (NOTSET, 97),
    ]

    def __init__(self):
        StreamHandler.__init__(self, sys.stderr)

    def format(self, record: logging.LogRecord) -> str:
        s = StreamHandler.format(self, record)
        for level, color in self.COLOR_LEVELS:
            if record.levelno >= level:
                return self._color(s, color)

    def _color(self, s: str, color: int) -> str:
        return '\033[%dm%s\033[0m' % (color, s)



from contextlib import suppress
from typing import Any, Type, Optional

from betty.ancestry import Person, File, Place, Identifiable, PersonName, IdentifiableSource, IdentifiableEvent, \
    IdentifiableCitation
from betty.config import Configuration
from betty.media_type import EXTENSIONS


class LocalizedUrlGenerator:
    def generate(self, resource: Any, media_type: str, absolute: bool = False, locale: Optional[str] = None) -> str:
        raise NotImplementedError


class StaticUrlGenerator:
    def generate(self, resource: Any, absolute: bool = False, ) -> str:
        raise NotImplementedError


class LocalizedPathUrlGenerator(LocalizedUrlGenerator):
    def __init__(self, configuration: Configuration):
        self._configuration = configuration

    def generate(self, resource, media_type, **kwargs) -> str:
        return _generate_from_path(self._configuration, resource, localize=True, **kwargs)


class StaticPathUrlGenerator(StaticUrlGenerator):
    def __init__(self, configuration: Configuration):
        self._configuration = configuration

    def generate(self, resource, **kwargs) -> str:
        return _generate_from_path(self._configuration, resource, localize=False, **kwargs)


class IdentifiableResourceUrlGenerator(LocalizedUrlGenerator):
    def __init__(self, configuration: Configuration, identifiable_type: Type[Identifiable], pattern: str):
        self._configuration = configuration
        self._type = identifiable_type
        self._pattern = pattern

    def generate(self, resource: Identifiable, media_type, **kwargs) -> str:
        if not isinstance(resource, self._type):
            raise ValueError('%s is not a %s' % (type(resource), self._type))
        kwargs['localize'] = True
        return _generate_from_path(self._configuration, self._pattern % (resource.id, EXTENSIONS[media_type]), **kwargs)


class PersonNameUrlGenerator(LocalizedUrlGenerator):
    def __init__(self, person_url_generator: LocalizedUrlGenerator):
        self._person_url_generator = person_url_generator

    def generate(self, name: PersonName, *args, **kwargs) -> str:
        if not isinstance(name, PersonName):
            raise ValueError('%s is not a %s' % (type(name), PersonName))
        return self._person_url_generator.generate(name.person, *args, **kwargs)


class SiteUrlGenerator(LocalizedUrlGenerator):
    def __init__(self, configuration: Configuration):
        person_url_generator = IdentifiableResourceUrlGenerator(configuration, Person, 'person/%s/index.%s')
        self._generators = [
            person_url_generator,
            PersonNameUrlGenerator(person_url_generator),
            IdentifiableResourceUrlGenerator(
                configuration, IdentifiableEvent, 'event/%s/index.%s'),
            IdentifiableResourceUrlGenerator(
                configuration, Place, 'place/%s/index.%s'),
            IdentifiableResourceUrlGenerator(
                configuration, File, 'file/%s/index.%s'),
            IdentifiableResourceUrlGenerator(
                configuration, IdentifiableSource, 'source/%s/index.%s'),
            IdentifiableResourceUrlGenerator(
                configuration, IdentifiableCitation, 'citation/%s/index.%s'),
            LocalizedPathUrlGenerator(configuration),
        ]

    def generate(self, resource: Any, *args, **kwargs) -> str:
        for generator in self._generators:
            with suppress(ValueError):
                return generator.generate(resource, *args, **kwargs)
        raise ValueError('No URL generator found for %s.' % (
            resource if isinstance(resource, str) else type(resource)))


def _generate_from_path(configuration: Configuration, path: str, localize: bool = False, absolute: bool = False, locale: Optional[str] = None) -> str:
    if not isinstance(path, str):
        raise ValueError('%s is not a string.' % type(path))
    url = configuration.base_url if absolute else ''
    url += configuration.root_path
    if localize and configuration.multilingual:
        if locale is None:
            locale = configuration.default_locale
        url += configuration.locales[locale].alias + '/'
    url += path.lstrip('/')
    if configuration.clean_urls and (path.endswith('/index.html') or path == 'index.html'):
        url = url[:-10]
    return url



from betty.ancestry import Person, Presence, Event, PersonName, Source, Citation, Birth, Subject, Death
from betty.asyncio import sync
from betty.locale import Date
from betty.tests import TemplateTestCase


class Test(TemplateTestCase):
    template_file = 'meta/person.html.j2'

    @sync
    async def test_without_meta(self):
        person = Person('P0')
        expected = '<div class="meta"></div>'
        async with self._render(data={
            'person': person,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_private(self):
        person = Person('P0')
        person.private = True
        expected = '<div class="meta"><p>This person\'s details are unavailable to protect their privacy.</p></div>'
        async with self._render(data={
            'person': person,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_with_one_alternative_name(self):
        person = Person('P0')
        person.names.append(PersonName('Jane', 'Dough'))
        name = PersonName('Janet', 'Doughnut')
        name.citations.append(Citation(Source('The Source')))
        person.names.append(name)
        expected = '<div class="meta"><span class="aka">Also known as <span class="person-label" typeof="foaf:Person"><span property="foaf:individualName">Janet</span> <span property="foaf:familyName">Doughnut</span></span><a href="#reference-1" class="citation">[1]</a></span></div>'
        async with self._render(data={
            'person': person,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_with_multiple_alternative_names(self):
        person = Person('P0')
        person.names.append(PersonName('Jane', 'Dough'))
        person.names.append(PersonName('Janet', 'Doughnut'))
        person.names.append(PersonName('Janetar', 'Of Doughnuton'))
        expected = '<div class="meta"><span class="aka">Also known as <span class="person-label" typeof="foaf:Person"><span property="foaf:individualName">Janet</span> <span property="foaf:familyName">Doughnut</span></span>, <span class="person-label" typeof="foaf:Person"><span property="foaf:individualName">Janetar</span> <span property="foaf:familyName">Of Doughnuton</span></span></span></div>'
        async with self._render(data={
            'person': person,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_with_start(self):
        person = Person('P0')
        Presence(person, Subject(), Event(Birth(), Date(1970)))
        expected = '<div class="meta"><dl><dt>Birth</dt><dd>1970</dd></dl></div>'
        async with self._render(data={
            'person': person,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_with_end(self):
        person = Person('P0')
        Presence(person, Subject(), Event(Death(), Date(1970)))
        expected = '<div class="meta"><dl><dt>Death</dt><dd>1970</dd></dl></div>'
        async with self._render(data={
            'person': person,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_embedded(self):
        person = Person('P0')
        Presence(person, Subject(), Event(Birth(), Date(1970)))
        person.names.append(PersonName('Jane', 'Dough'))
        name = PersonName('Janet', 'Doughnut')
        name.citations.append(Citation(Source('The Source')))
        person.names.append(name)
        expected = '<div class="meta"><span class="aka">Also known as <span class="person-label" typeof="foaf:Person"><span property="foaf:individualName">Janet</span> <span property="foaf:familyName">Doughnut</span></span></span><dl><dt>Birth</dt><dd>1970</dd></dl></div>'
        async with self._render(data={
            'person': person,
            'embedded': True,
        }) as (actual, _):
            self.assertEqual(expected, actual)



import json as stdjson
from os import path
from tempfile import TemporaryDirectory

import jsonschema
from parameterized import parameterized

from betty.config import Configuration
from betty.asyncio import sync
from betty.openapi import build_specification
from betty.site import Site
from betty.tests import TestCase


class BuildSpecificationTest(TestCase):
    @parameterized.expand([
        (True,),
        (False,),
    ])
    @sync
    async def test(self, content_negotiation: str):
        with open(path.join(path.dirname(__file__), 'test_openapi_assets', 'schema.json')) as f:
            schema = stdjson.load(f)
        with TemporaryDirectory() as output_directory_path:
            configuration = Configuration(
                output_directory_path, 'https://example.com')
            configuration.content_negotiation = content_negotiation
            async with Site(configuration) as site:
                specification = build_specification(site)
        jsonschema.validate(specification, schema)



import hashlib
import logging
import shutil
from contextlib import suppress
from os import path
from os.path import dirname
from typing import Optional, Iterable, Any

from betty import subprocess
from betty.fs import DirectoryBackup
from betty.generate import PostStaticGenerator
from betty.html import HtmlProvider
from betty.plugin import Plugin, NO_CONFIGURATION
from betty.site import Site


class Trees(Plugin, HtmlProvider, PostStaticGenerator):
    def __init__(self, site: Site):
        self._site = site

    @classmethod
    def for_site(cls, site: Site, configuration: Any = NO_CONFIGURATION):
        return cls(site)

    async def post_static_generate(self) -> None:
        await self._render()

    @property
    def assets_directory_path(self) -> Optional[str]:
        return '%s/assets' % dirname(__file__)

    async def _render(self) -> None:
        build_directory_path = path.join(self._site.configuration.cache_directory_path, self.name(),
                                         hashlib.md5(self.assets_directory_path.encode()).hexdigest(), 'build')

        async with DirectoryBackup(build_directory_path, 'node_modules'):
            with suppress(FileNotFoundError):
                shutil.rmtree(build_directory_path)
            shutil.copytree(path.join(self.assets_directory_path, 'js'), build_directory_path)
        await self._site.renderer.render_tree(build_directory_path)

        self._site.executor.submit(_do_render, build_directory_path, self._site.configuration.www_directory_path)

    @property
    def public_css_paths(self) -> Iterable[str]:
        return {
            self._site.static_url_generator.generate('trees.css'),
        }

    @property
    def public_js_paths(self) -> Iterable[str]:
        return {
            self._site.static_url_generator.generate('trees.js'),
        }


def _do_render(build_directory_path: str, www_directory_path: str) -> None:
    # Install third-party dependencies.
    subprocess.run(['npm', 'install', '--production'], cwd=build_directory_path)

    # Run Webpack.
    subprocess.run(['npm', 'run', 'webpack'], cwd=build_directory_path)
    output_directory_path = path.join(path.dirname(build_directory_path), 'output')
    shutil.copy2(path.join(output_directory_path, 'trees.css'), path.join(www_directory_path, 'trees.css'))
    shutil.copy2(path.join(output_directory_path, 'trees.js'), path.join(www_directory_path, 'trees.js'))

    logging.getLogger().info('Built the interactive family trees.')



"""Integrates Betty with Python's setuptools."""

import os
from glob import glob
from os.path import abspath, dirname, join

from setuptools import setup, find_packages

ROOT_PATH = os.path.dirname(os.path.abspath(__file__))

with open('/'.join((ROOT_PATH, 'VERSION'))) as f:
    VERSION = f.read()

with open('/'.join((ROOT_PATH, 'README.md'))) as f:
    long_description = f.read()

SETUP = {
    'name': 'betty',
    'description': 'Betty is a static ancestry site generator.',
    'long_description': long_description,
    'long_description_content_type': 'text/markdown',
    'version': VERSION,
    'license': 'GPLv3',
    'author': 'Bart Feenstra & contributors',
    'author_email': 'bart@mynameisbart.com',
    'url': 'https://github.com/bartfeenstra/betty',
    'classifiers': [
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Operating System :: MacOS',
        'Operating System :: POSIX :: BSD',
        'Operating System :: POSIX :: Linux',
        'Operating System :: POSIX :: SunOS/Solaris',
        'Operating System :: Unix',
        'Topic :: Scientific/Engineering :: Visualization',
        'Topic :: Sociology :: Genealogy',
        'Topic :: Software Development :: Code Generators',
    ],
    'python_requires': '~= 3.6',
    'install_requires': [
        'aiohttp ~= 3.7.3',
        'async-exit-stack ~= 1.0.1; python_version <= "3.6"',
        'async_generator ~= 1.10; python_version <= "3.6"',
        'babel ~= 2.9.0',
        'click ~= 7.1.2',
        'docker ~= 4.3.1',
        'geopy ~= 2.0.0',
        'jinja2 ~= 2.11.2',
        'jsonschema ~= 3.2.0',
        'libsass ~= 0.20.1',
        'lxml ~= 4.6.1',
        'markupsafe ~= 1.1.1',
        'pdf2image ~= 1.14.0 ',
        'python-resize-image ~= 1.1.19',
        'pyyaml ~= 5.3.1',
        'voluptuous ~= 0.12.0',
    ],
    'extras_require': {
        'development': [
            'aioresponses ~= 0.7.1',
            'autopep8 ~= 1.5.4',
            'codecov ~= 2.1.10',
            'coverage ~= 5.3',
            'flake8 ~= 3.7.0',
            'html5lib ~= 1.1',
            'mock ~= 4.0.2; python_version <= "3.7"',
            'nose2 ~= 0.9.2',
            'parameterized ~= 0.7.4',
            'setuptools ~= 50.3.2',
            'twine ~= 3.2.0',
            'wheel ~= 0.35.1',
        ],
    },
    'entry_points': {
        'console_scripts': [
            'betty=betty.cli:main',
        ],
    },
    'packages': find_packages(),
    'data_files': [
        ('', [
            'LICENSE.txt',
            'README.md',
            'VERSION',
        ])
    ],
    'include_package_data': True,
    'package_data': {
        'betty': glob(join(dirname(abspath(__file__)), 'betty', 'assets', '**'), recursive=True) + glob(join(dirname(abspath(__file__)), 'betty', 'plugin', '*', 'assets', '**'), recursive=True),
    },
}

if __name__ == '__main__':
    setup(**SETUP)



from betty.asyncio import sync
from betty.html import HtmlProvider
from betty.tests import TestCase


class HtmlProviderTest(TestCase):
    class _HtmlProviderDummy(HtmlProvider):
        pass

    @sync
    async def test_public_css_paths(self):
        sut = self._HtmlProviderDummy()
        self.assertEquals(0, len(sut.public_css_paths))

    @sync
    async def test_public_js_paths(self):
        sut = self._HtmlProviderDummy()
        self.assertEquals(0, len(sut.public_js_paths))



from contextlib import suppress

import docker
from docker.errors import NotFound
from docker.models.containers import Container as DockerContainer


class Container:
    def __init__(self, www_directory_path: str, docker_directory_path: str, nginx_configuration_file_path: str, name: str):
        self._name = name
        self._docker_directory_path = docker_directory_path
        self._nginx_configuration_file_path = nginx_configuration_file_path
        self._www_directory_path = www_directory_path
        self._client = docker.from_env()

    def __enter__(self) -> None:
        self.start()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    def start(self) -> None:
        # Stop any containers that may have been left over.
        self.stop()

        self._client.images.build(path=self._docker_directory_path, tag=self._name)
        self._client.containers.run(self._name, name=self._name, auto_remove=True, detach=True, volumes={
            self._nginx_configuration_file_path: {
                'bind': '/etc/nginx/conf.d/betty.conf',
                'mode': 'ro',
            },
            self._www_directory_path: {
                'bind': '/var/www/betty',
                'mode': 'ro',
            },
        })
        self._container.exec_run(['nginx', '-s', 'reload'])

    def stop(self) -> None:
        with suppress(NotFound):
            self._container.stop()

    @property
    def _container(self) -> DockerContainer:
        return self._client.containers.get(self._name)

    @property
    def ip(self) -> DockerContainer:
        return self._client.api.inspect_container(self._name)['NetworkSettings']['Networks']['bridge']['IPAddress']



from betty.ancestry import Person, PersonName
from betty.asyncio import sync
from betty.tests import TemplateTestCase


class Test(TemplateTestCase):
    template_file = 'label/person.html.j2'

    @sync
    async def test_with_name(self):
        person = Person('P0')
        person.names.append(PersonName('Jane', 'Dough'))
        expected = '<a href="/person/P0/index.html"><span class="person-label" typeof="foaf:Person"><span property="foaf:individualName">Jane</span> <span property="foaf:familyName">Dough</span></span></a>'
        async with self._render(data={
            'person': person,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_without_name(self):
        person = Person('P0')
        expected = '<a href="/person/P0/index.html"><span class="nn" title="This person\'s name is unknown.">n.n.</span></a>'
        async with self._render(data={
            'person': person,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_embedded(self):
        person = Person('P0')
        expected = '<span class="nn" title="This person\'s name is unknown.">n.n.</span>'
        async with self._render(data={
            'person': person,
            'embedded': True,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_person_is_context(self):
        person = Person('P0')
        expected = '<span class="nn" title="This person\'s name is unknown.">n.n.</span>'
        async with self._render(data={
            'person': person,
            'person_context': person,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_private(self):
        person = Person('P0')
        person.private = True
        expected = '<a href="/person/P0/index.html"><span class="private" title="This person\'s details are unavailable to protect their privacy.">private</span></a>'
        async with self._render(data={
            'person': person,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_with_resource(self):
        person = Person('P0')
        person.names.append(PersonName('Jane', 'Dough'))
        expected = '<a href="/person/P0/index.html"><span class="person-label" typeof="foaf:Person"><span property="foaf:individualName">Jane</span> <span property="foaf:familyName">Dough</span></span></a>'
        async with self._render(data={
            'resource': person,
        }) as (actual, _):
            self.assertEqual(expected, actual)



from unittest.mock import Mock

from betty.plugin import Plugin
from betty.site import Site
from betty.tests import TestCase


class PluginTest(TestCase):
    def test_for_site(self):
        site = Mock(Site)
        plugin = Plugin.for_site(site)
        self.assertIsInstance(plugin, Plugin)

    def test_depends_on(self):
        self.assertEquals(set(), Plugin.depends_on())



from os import makedirs, path
from tempfile import TemporaryDirectory
from typing import List, Dict, Optional, Iterable, Type
from unittest import TestCase
from unittest.mock import Mock

from parameterized import parameterized

from betty.ancestry import File, PlaceName, Subject, Attendee, Witness, Dated, Resource, Person, Place, Citation
from betty.config import Configuration, LocaleConfiguration
from betty.asyncio import sync
from betty.jinja2 import Jinja2Renderer, _Citer, Jinja2Provider
from betty.locale import Date, Datey, DateRange, Localized
from betty.media_type import MediaType
from betty.plugin import Plugin
from betty.site import Site
from betty.tests import TemplateTestCase


class Jinja2ProviderTest(TestCase):
    def test_globals(self) -> None:
        sut = Jinja2Provider()
        self.assertIsInstance(sut.globals, dict)

    def test_filters(self) -> None:
        sut = Jinja2Provider()
        self.assertIsInstance(sut.filters, dict)


class Jinja2RendererTest(TestCase):
    @sync
    async def test_render_file(self) -> None:
        with TemporaryDirectory() as output_directory_path:
            configuration = Configuration(output_directory_path, 'https://ancestry.example.com')
            async with Site(configuration) as site:
                sut = Jinja2Renderer(site.jinja2_environment, configuration)
                template = '{% if true %}true{% endif %}'
                expected_output = 'true'
                with TemporaryDirectory() as working_directory_path:
                    template_file_path = path.join(working_directory_path, 'betty.txt.j2')
                    with open(template_file_path, 'w') as f:
                        f.write(template)
                    await sut.render_file(template_file_path)
                    with open(path.join(working_directory_path, 'betty.txt')) as f:
                        self.assertEquals(expected_output, f.read().strip())
                    self.assertFalse(path.exists(template_file_path))

    @sync
    async def test_render_file_should_ignore_non_sass_or_scss(self) -> None:
        with TemporaryDirectory() as output_directory_path:
            configuration = Configuration(output_directory_path, 'https://ancestry.example.com')
            async with Site(configuration) as site:
                sut = Jinja2Renderer(site.jinja2_environment, configuration)
                template = '{% if true %}true{% endif %}'
                with TemporaryDirectory() as working_directory_path:
                    template_file_path = path.join(working_directory_path, 'betty.txt')
                    with open(template_file_path, 'w') as f:
                        f.write(template)
                    await sut.render_file(template_file_path)
                    with open(path.join(working_directory_path, 'betty.txt')) as f:
                        self.assertEquals(template, f.read())

    @sync
    async def test_render_tree(self) -> None:
        with TemporaryDirectory() as output_directory_path:
            configuration = Configuration(output_directory_path, 'https://ancestry.example.com')
            async with Site(configuration) as site:
                sut = Jinja2Renderer(site.jinja2_environment, configuration)
                template = '{% if true %}true{% endif %}'
                expected_output = 'true'
                with TemporaryDirectory() as working_directory_path:
                    working_subdirectory_path = path.join(working_directory_path, 'sub')
                    makedirs(working_subdirectory_path)
                    scss_file_path = path.join(working_subdirectory_path, 'betty.txt.j2')
                    with open(scss_file_path, 'w') as f:
                        f.write(template)
                    await sut.render_tree(working_directory_path)
                    with open(path.join(working_subdirectory_path, 'betty.txt')) as f:
                        self.assertEquals(expected_output, f.read().strip())
                    self.assertFalse(path.exists(scss_file_path))


class FilterFlattenTest(TemplateTestCase):
    @parameterized.expand([
        ('', '{{ [] | flatten | join(", ") }}'),
        ('', '{{ [[], [], []] | flatten | join(", ") }}'),
        ('kiwi, apple, banana',
         '{{ [["kiwi"], ["apple"], ["banana"]] | flatten | join(", ") }}'),
    ])
    @sync
    async def test(self, expected, template):
        async with self._render(template_string=template) as (actual, _):
            self.assertEquals(expected, actual)


class FilterWalkTest(TemplateTestCase):
    class WalkData:
        def __init__(self, label, children=None):
            self._label = label
            self.children = children or []

        def __str__(self):
            return self._label

    @parameterized.expand([
        ('', '{{ data | walk("children") | join }}', WalkData('parent')),
        ('child1, child1child1, child2', '{{ data | walk("children") | join(", ") }}',
         WalkData('parent', [WalkData('child1', [WalkData('child1child1')]), WalkData('child2')])),
    ])
    @sync
    async def test(self, expected, template, data):
        async with self._render(template_string=template, data={
            'data': data,
        }) as (actual, _):
            self.assertEquals(expected, actual)


class FilterParagraphsTest(TemplateTestCase):
    @parameterized.expand([
        ('<p></p>', '{{ "" | paragraphs }}'),
        ('<p>Apples <br>\n and <br>\n oranges</p>',
         '{{ "Apples \n and \n oranges" | paragraphs }}'),
    ])
    @sync
    async def test(self, expected, template):
        async with self._render(template_string=template) as (actual, _):
            self.assertEquals(expected, actual)


class FilterFormatDegreesTest(TemplateTestCase):
    @parameterized.expand([
        ('0° 0&#39; 0&#34;', '{{ 0 | format_degrees }}'),
        ('52° 22&#39; 1&#34;', '{{ 52.367 | format_degrees }}'),
    ])
    @sync
    async def test(self, expected, template):
        async with self._render(template_string=template) as (actual, _):
            self.assertEquals(expected, actual)


class FilterMapTest(TemplateTestCase):
    class MapData:
        def __init__(self, label):
            self.label = label

    @parameterized.expand([
        ('kiwi, apple, banana', '{{ data | map(attribute="label") | join(", ") }}',
         [MapData('kiwi'), MapData('apple'), MapData('banana')]),
        ('kiwi, None, apple, None, banana',
         '{% macro print_string(value) %}{% if value is none %}None{% else %}{{ value }}{% endif %}{% endmacro %}{{ ["kiwi", None, "apple", None, "banana"] | map(print_string) | join(", ") }}',
         {}),
    ])
    @sync
    async def test(self, expected, template, data):
        async with self._render(template_string=template, data={
            'data': data,
        }) as (actual, _):
            self.assertEquals(expected, actual)


class FilterFileTest(TemplateTestCase):
    @parameterized.expand([
        ('/file/F1.py', '{{ file | file }}', File('F1', __file__)),
        ('/file/F1.py:/file/F1.py',
         '{{ file | file }}:{{ file | file }}', File('F1', __file__)),
    ])
    @sync
    async def test(self, expected, template, file):
        async with self._render(template_string=template, data={
            'file': file,
        }) as (actual, site):
            self.assertEquals(expected, actual)
            for file_path in actual.split(':'):
                self.assertTrue(path.exists(path.join(site.configuration.www_directory_path, file_path[1:])))


class FilterImageTest(TemplateTestCase):
    image_path = path.join(path.dirname(path.dirname(__file__)), 'assets', 'public', 'static', 'betty-512x512.png')

    @parameterized.expand([
        ('/file/F1-99x-.png',
         '{{ file | image(width=99) }}', File('F1', image_path, media_type=MediaType('image/png'))),
        ('/file/F1--x99.png',
         '{{ file | image(height=99) }}', File('F1', image_path, media_type=MediaType('image/png'))),
        ('/file/F1-99x99.png',
         '{{ file | image(width=99, height=99) }}', File('F1', image_path, media_type=MediaType('image/png'))),
        ('/file/F1-99x99.png:/file/F1-99x99.png',
         '{{ file | image(width=99, height=99) }}:{{ file | image(width=99, height=99) }}', File('F1', image_path, media_type=MediaType('image/png'))),
    ])
    @sync
    async def test(self, expected, template, file):
        async with self._render(template_string=template, data={
            'file': file,
        }) as (actual, site):
            self.assertEquals(expected, actual)
            for file_path in actual.split(':'):
                self.assertTrue(path.exists(path.join(site.configuration.www_directory_path, file_path[1:])))

    @sync
    async def test_without_width(self):
        file = File('F1', self.image_path, media_type=MediaType('image/png'))
        with self.assertRaises(ValueError):
            async with self._render(template_string='{{ file | image }}', data={
                'file': file,
            }):
                pass


class TestPlugin(Plugin):
    """
    This class must be top-level. Otherwise it cannot be imported by its fully qualified name.
    """
    pass


class GlobalPluginsTest(TemplateTestCase):
    @sync
    async def test_getitem_with_unknown_plugin(self):
        template = '{{ plugins["betty.UnknownPlugin"] | default("false") }}'
        async with self._render(template_string=template) as (actual, _):
            self.assertEquals('false', actual)

    @sync
    async def test_getitem_with_disabled_plugin(self):
        template = '{{ plugins["%s"] | default("false") }}' % TestPlugin.name()
        async with self._render(template_string=template) as (actual, _):
            self.assertEquals('false', actual)

    @sync
    async def test_getitem_with_enabled_plugin(self):
        template = '{%% if plugins["%s"] is not none %%}true{%% else %%}false{%% endif %%}' % TestPlugin.name()

        def _update_configuration(configuration: Configuration) -> None:
            configuration.plugins[TestPlugin] = None
        async with self._render(template_string=template, update_configuration=_update_configuration) as (actual, _):
            self.assertEquals('true', actual)

    @sync
    async def test_contains_with_unknown_plugin(self):
        template = '{% if "betty.UnknownPlugin" in plugins %}true{% else %}false{% endif %}'
        async with self._render(template_string=template) as (actual, _):
            self.assertEquals('false', actual)

    @sync
    async def test_contains_with_disabled_plugin(self):
        template = '{%% if "%s" in plugins %%}true{%% else %%}false{%% endif %%}' % TestPlugin.name()
        async with self._render(template_string=template) as (actual, _):
            self.assertEquals('false', actual)

    @sync
    async def test_contains_with_enabled_plugin(self):
        template = '{%% if "%s" in plugins %%}true{%% else %%}false{%% endif %%}' % TestPlugin.name()

        def _update_configuration(configuration: Configuration) -> None:
            configuration.plugins[TestPlugin] = None
        async with self._render(template_string=template, update_configuration=_update_configuration) as (actual, _):
            self.assertEquals('true', actual)


class GlobalCiterTest(TemplateTestCase):
    @sync
    async def test_cite(self):
        citation1 = Mock(Citation)
        citation2 = Mock(Citation)
        sut = _Citer()
        self.assertEquals(1, sut.cite(citation1))
        self.assertEquals(2, sut.cite(citation2))
        self.assertEquals(1, sut.cite(citation1))

    @sync
    async def test_iter(self):
        citation1 = Mock(Citation)
        citation2 = Mock(Citation)
        sut = _Citer()
        sut.cite(citation1)
        sut.cite(citation2)
        self.assertEquals([(1, citation1), (2, citation2)], list(sut))

    @sync
    async def test_len(self):
        citation1 = Mock(Citation)
        citation2 = Mock(Citation)
        sut = _Citer()
        sut.cite(citation1)
        sut.cite(citation2)
        self.assertEquals(2, len(sut))


class FormatDateTest(TemplateTestCase):
    @sync
    async def test(self):
        template = '{{ date | format_date }}'
        date = Date(1970, 1, 1)
        async with self._render(template_string=template, data={
            'date': date,
        }) as (actual, _):
            self.assertEquals('January 1, 1970', actual)


class FilterSortLocalizedsTest(TemplateTestCase):
    class WithLocalizedNames:
        def __init__(self, identifier, names: List[PlaceName]):
            self.id = identifier
            self.names = names

        def __repr__(self):
            return self.id

    @sync
    async def test(self):
        template = '{{ data | sort_localizeds(localized_attribute="names", sort_attribute="name") }}'
        data = [
            self.WithLocalizedNames('third', [
                PlaceName('3', 'nl-NL'),
            ]),
            self.WithLocalizedNames('second', [
                PlaceName('2', 'en'),
                PlaceName('1', 'nl-NL'),
            ]),
            self.WithLocalizedNames('first', [
                PlaceName('2', 'nl-NL'),
                PlaceName('1', 'en-US'),
            ]),
        ]
        async with self._render(template_string=template, data={
            'data': data,
        }) as (actual, _):
            self.assertEquals('[first, second, third]', actual)

    @sync
    async def test_with_empty_iterable(self):
        template = '{{ data | sort_localizeds(localized_attribute="names", sort_attribute="name") }}'
        async with self._render(template_string=template, data={
            'data': [],
        }) as (actual, _):
            self.assertEquals('[]', actual)


class FilterSelectLocalizedsTest(TemplateTestCase):
    @parameterized.expand([
        ('', 'en', []),
        ('Apple', 'en', [
            PlaceName('Apple', 'en')
        ]),
        ('Apple', 'en', [
            PlaceName('Apple', 'en-US')
        ]),
        ('Apple', 'en-US', [
            PlaceName('Apple', 'en')
        ]),
        ('', 'nl', [
            PlaceName('Apple', 'en')
        ]),
        ('', 'nl-NL', [
            PlaceName('Apple', 'en')
        ]),
    ])
    @sync
    async def test(self, expected: str, locale: str, data: Iterable[Localized]):
        template = '{{ data | select_localizeds | map(attribute="name") | join(", ") }}'

        def _update_configuration(configuration: Configuration) -> None:
            configuration.locales.clear()
            configuration.locales[locale] = LocaleConfiguration(locale)
        async with self._render(template_string=template, data={
            'data': data,
        }, update_configuration=_update_configuration) as (actual, _):
            self.assertEquals(expected, actual)


class FilterSelectDatedsTest(TemplateTestCase):
    class DatedDummy(Dated):
        def __init__(self, value: str, date: Optional[Datey] = None):
            Dated.__init__(self)
            self._value = value
            self.date = date

        def __str__(self):
            return self._value

    @parameterized.expand([
        ('Apple', {
            'dateds': [
                DatedDummy('Apple'),
            ],
            'date': None,
        }),
        ('Apple', {
            'dateds': [
                DatedDummy('Apple'),
            ],
            'date': Date(),
        }),
        ('Apple', {
            'dateds': [
                DatedDummy('Apple'),
            ],
            'date': Date(1970, 1, 1),
        }),
        ('', {
            'dateds': [
                DatedDummy('Apple', Date(1970, 1, 1)),
            ],
            'date': None,
        }),
        ('', {
            'dateds': [
                DatedDummy('Apple', Date(1970, 1, 1)),
            ],
            'date': Date(),
        }),
        ('Apple', {
            'dateds': [
                DatedDummy('Apple', Date(1970, 1, 1)),
            ],
            'date': Date(1970, 1, 1),
        }),
        ('Apple, Strawberry', {
            'dateds': [
                DatedDummy('Apple', Date(1971, 1, 1)),
                DatedDummy('Strawberry', Date(1970, 1, 1)),
                DatedDummy('Banana', Date(1969, 1, 1)),
                DatedDummy('Orange', Date(1972, 12, 31)),
            ],
            'date': DateRange(Date(1970, 1, 1), Date(1971, 1, 1)),
        }),
    ])
    @sync
    async def test(self, expected: str, data: Dict):
        template = '{{ dateds | select_dateds(date=date) | join(", ") }}'
        async with self._render(template_string=template, data=data) as (actual, _):
            self.assertEquals(expected, actual)


class TestSubjectRoleTest(TemplateTestCase):
    @parameterized.expand([
        ('true', Subject()),
        ('false', Subject),
        ('false', Attendee()),
        ('false', 9),
    ])
    @sync
    async def test(self, expected, data) -> None:
        template = '{% if data is subject_role %}true{% else %}false{% endif %}'
        async with self._render(template_string=template, data={
            'data': data,
        }) as (actual, _):
            self.assertEquals(expected, actual)


class TestWitnessRoleTest(TemplateTestCase):
    @parameterized.expand([
        ('true', Witness()),
        ('false', Witness),
        ('false', Attendee()),
        ('false', 9),
    ])
    @sync
    async def test(self, expected, data) -> None:
        template = '{% if data is witness_role %}true{% else %}false{% endif %}'
        async with self._render(template_string=template, data={
            'data': data,
        }) as (actual, _):
            self.assertEquals(expected, actual)


class TestResourceTest(TemplateTestCase):
    @parameterized.expand([
        ('true', Person, Person('P1')),
        ('false', Person, Place('P1', [PlaceName('The Place')])),
        ('true', Place, Place('P1', [PlaceName('The Place')])),
        ('false', Place, Person('P1')),
        ('false', Place, 999),
        ('false', Person, object()),
    ])
    @sync
    async def test(self, expected, resource_type: Type[Resource], data) -> None:
        template = f'{{% if data is {resource_type.resource_type_name()}_resource %}}true{{% else %}}false{{% endif %}}'
        async with self._render(template_string=template, data={
            'data': data,
        }) as (actual, _):
            self.assertEquals(expected, actual)
